import os
import pickle
import numpy as np
from tensorflow.keras.utils import Sequence
from tensorflow.keras.layers import (Dense, Input)
from tensorflow.keras.layers import (Activation, Dense, Input, Conv1D, Conv2D, MaxPooling2D, Reshape,
                                     Dropout, SpatialDropout1D, SpatialDropout2D)
from tensorflow.keras.models import Sequential, Model
import madmom

import tensorflow.keras.backend as K
import tensorflow as tf

import tensorflow_addons as tfa

from modules.utils import PKL_PATH

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "10"
os.environ["TF_DUMP_GRAPH_PREFIX"] = 'tmp'


# GENERAL CONSTANTS
FPS = 100  # set the frame rate as FPS frames per second
MASK_VALUE = -1

lr = 0.05
num_epochs = 50
dropout_rate = 0.15
num_filters = 16
num_dilations = 11
kernel_size = 5
activation = 'elu'


class Fold(object):

    def __init__(self, folds, fold):
        self.folds = folds
        self.fold = fold

    @property
    def test(self):
        # fold N for testing
        return np.unique(self.folds[self.fold])

    @property
    def val(self):
        # fold N+1 for validation
        return np.unique(self.folds[(self.fold + 1) % len(self.folds)])

    @property
    def train(self):
        # all remaining folds for training
        train = np.hstack(self.folds)
        train = np.setdiff1d(train, self.val)
        train = np.setdiff1d(train, self.test)
        return train


def cnn_pad(data, pad_frames):
    """Pad the data by repeating the first and last frame N times."""
    pad_start = np.repeat(data[:1], pad_frames, axis=0)
    pad_stop = np.repeat(data[-1:], pad_frames, axis=0)
    return np.concatenate((pad_start, data, pad_stop))


def residual_block_v1(x, dilation_rate, activation, num_filters, kernel_size, padding, dropout_rate=0, name=''):
    original_x = x
    conv = Conv1D(num_filters, kernel_size=kernel_size,
                  dilation_rate=dilation_rate, padding='same',
                  name=name + '_%d_dilated_conv' % (dilation_rate))(x)
    x = Activation(activation, name=name + '_%d_activation' % (dilation_rate))(conv)
    x = SpatialDropout1D(dropout_rate, name=name + '_%d_spatial_dropout_%.2f' % (dilation_rate, dropout_rate))(x)
    x = Conv1D(num_filters, 1, padding='same', name=name + '_%d_conv_1x1' % (dilation_rate))(x)
    res_x = tf.keras.layers.add([original_x, x], name=name + '_%d_residual' % (dilation_rate))
    return res_x, x


class TCN_v1():
    def __init__(self, num_filters=8, kernel_size=5, dilations=[1, 2, 4, 8, 16, 32, 64, 128],
                 activation='elu', dropout_rate=0.15, name='tcn'):
        self.name = name
        self.dropout_rate = dropout_rate
        self.activation = activation
        self.dilations = dilations
        self.kernel_size = kernel_size
        self.num_filters = num_filters

    def __call__(self, inputs):
        x = inputs
        for d in self.dilations:
            x, _ = residual_block_v1(x, d, self.activation, self.num_filters,
                                     self.kernel_size, self.dropout_rate, name=self.name)
        x = Activation(self.activation, name=self.name + '_activation')(x)
        return x


class DataSequence_TCNv1(Sequence):

    def __init__(self, x, y, fps=FPS, pad_frames=None):
        self.x = x
        self.y = [madmom.utils.quantize_events(o, fps=fps, length=len(d))
                  for o, d in zip(y, self.x)]
        self.pad_frames = pad_frames

    def __len__(self):
        return len(self.x)

    def __getitem__(self, idx):
        x = np.array(cnn_pad(self.x[idx], self.pad_frames))[np.newaxis, ..., np.newaxis]
        y = self.y[idx][np.newaxis, ..., np.newaxis]
        return x, y

    def widen_targets(self, size=3, value=0.5):
        from scipy.ndimage import maximum_filter1d
        for y in self.y:
            np.maximum(y, maximum_filter1d(y, size=size) * value, out=y)


def simple_TCN(dataset):
    train_db = pickle.load(open('%s/%s.pkl' % (PKL_PATH, dataset), 'rb'))
    num_fold = 0
    fold = Fold(train_db.folds, num_fold)
    train = DataSequence_TCNv1([train_db.x[i] for i in fold.train],
                               [train_db.annotations[i] for i in fold.train],
                               pad_frames=2)
    val = DataSequence_TCNv1([train_db.x[i] for i in fold.val],
                             [train_db.annotations[i] for i in fold.val],
                             pad_frames=2)
    train.widen_targets()
    val.widen_targets()
    input_layer = Input(shape=((None, ) + train[0][0].shape[-2:]))
    conv_1 = Conv2D(num_filters, (3, 3), padding='valid',
                    name='conv_1_convolution')(input_layer)
    conv_1 = Activation(activation, name='conv_1_activation')(conv_1)
    conv_1 = MaxPooling2D((1, 3), name='conv_1_pooling')(conv_1)
    conv_1 = Dropout(dropout_rate, name='conv_1_dropout')(conv_1)
    conv_2 = Conv2D(num_filters, (3, 3), padding='valid',
                    name='conv_2_convolution')(conv_1)
    conv_2 = Activation(activation, name='conv_2_activation')(conv_2)
    conv_2 = MaxPooling2D((1, 3), name='conv_2_pooling')(conv_2)
    conv_2 = Dropout(dropout_rate, name='conv_2_dropout')(conv_2)
    conv_3 = Conv2D(num_filters, (1, 8), padding='valid',
                    name='conv_3_convolution')(conv_2)
    conv_3 = Activation(activation, name='conv_3_activation')(conv_3)
    conv_3 = Dropout(dropout_rate, name='conv_3_dropout')(conv_3)
    x = Reshape((-1, num_filters), name='tcn_reshape')(conv_3)
    dilations = [2 ** i for i in range(num_dilations)]
    tcn_layer = TCN_v1(num_filters=num_filters, kernel_size=kernel_size, dilations=dilations,
                       activation='elu', dropout_rate=dropout_rate, name='tcn')(x)
    output_layer = Dense(1, name='output')(tcn_layer)
    output_layer = Activation('sigmoid', name='output_activation')(output_layer)

    model = Model(input_layer, output_layer)

    radam = tfa.optimizers.RectifiedAdam(lr=lr, clipnorm=0.5)
    ranger = tfa.optimizers.Lookahead(radam, sync_period=6, slow_step_size=0.5)
    adam = tf.keras.optimizers.Adam()
    model.compile(optimizer=ranger, loss=K.binary_crossentropy, metrics=['binary_accuracy'])
    history = model.fit(train, steps_per_epoch=len(train), epochs=num_epochs, shuffle=True,
                        validation_data=val, validation_steps=len(val),
                        verbose=True)
    return True


if __name__ == "__main__":
    tf.config.set_soft_device_placement(True)
    dataset = 'traintest_smallsmc'
    simple_TCN(dataset)
